#ifndef OTC_DISPATCH_FAILURE_HH
#define OTC_DISPATCH_FAILURE_HH
/*
// ============================================================================
//
// = LIBRARY
//     OTC
// 
// = FILENAME
//     dispatch/failure.hh
//
// = AUTHOR(S)
//     Graham Dumpleton
// 
// = COPYRIGHT
//     Copyright 2000-2001 DUMPLETON SOFTWARE CONSULTING PTY LIMITED
//
// ============================================================================
*/

#include <OTC/dispatch/envelope.hh>
#include <OTC/dispatch/message.hh>

#ifdef __GNUG__
#pragma interface "OTC/dispatch/failure.hh"
#endif

/* ------------------------------------------------------------------------- */

class OSE_EXPORT OTCEV_Failure : public OTC_Event
    // = TITLE
    //     Event class encapsulating a failure response to a request.
    //
    // = DESCRIPTION
    //     The <OTCEV_Failure> class is the event delivered to a class
    //     associated with <OTC_SVBroker> in response to a prior request
    //     to another instance of <OTC_SVBroker> when a failure of some sort
    //     occurs. The <OTCEV_Failure> class contains the actual event sent
    //     from the other service and the instance of <OTCEV_Envelope> used
    //     to deliver it.
    //     
    // = NOTES
    //     No service descriptions are provided for either sender or receiver
    //     as either may be an anonymous agent which no service binding
    //     present. In a conversation involving multiple responses to a
    //     request, you should have remembered the conversation ID and save
    //     against that some knowledge of the participants in the
    //     conversation.
    //     
    // = SEE ALSO
    //     <OTC_SVBroker>, <OTCEV_Request>
{
  public:

			~OTCEV_Failure();

    // = INITIALISATION

			OTCEV_Failure(OTCEV_Envelope* theEnvelope);
				// <theEnvelope> should contain the event
				// received in failure response to the
				// request. Note that this class will call
				// <clone()> on <theEnvelope>. Thus, if
				// appropriate you should destroy the
				// original.

    // = QUERY

    OTCEV_Envelope*	envelope() const
				{ return envelope_; }
				// Returns the actual envelope event which
				// carried the event from the other service.

    OTC_String const&	recipient() const
				{ return envelope_->to(); }
				// Returns the agent address to which the
				// response as being delivered.

    OTC_String const&	sender() const
				{ return envelope_->from(); }
				// Returns the agent address from whom the
				// event was received.

    OTC_String const&	conversationId() const
				{ return envelope_->conversationId(); }
				// Returns the conversation ID sent with the
				// event. This should be the same as the
				// original request.

    int			sequenceNumber() const
				{ return envelope_->sequenceNumber(); }
				// Returns the sequence number sent with
				// the request this failure is being sent
				// in relation to.

    int			error() const
      				{ return error_; }
    				// Returns the error code identifying the
				// type of failure.

    OTC_String const&	description() const
    				{ return description_; }
				// Returns the description of the failure.

    OTC_String const&	origin() const
      				{ return origin_; }
				// Returns a string which identifies who
				// generated the error. This can also
				// act as a namespace qualifier such that
				// meaning can be applied to an error code.

    OTC_String const&	details() const
    				{ return details_; }
				// Returns additional details regarding
				// the failure.

    // = IDENTIFICATION

    void*               type() const;
                                // Returns a value that uniquely identifies
                                // this type of event.

    static void*        typeId()
                                { return &gTypeId_; }
                                // Returns a value that uniquely identifies
                                // this type of event.

    // = DEBUGGING

    void                dump(ostream& outs) const;
                                // Dumps info about the event onto <outs>.

  private:

			OTCEV_Failure(OTCEV_Failure const&);
				// Do not define an implementation for this.

    OTCEV_Failure&	operator=(OTCEV_Failure const&);
				// Do not define an implementation for this.

    static int          gTypeId_;
                                // Anchor for use as a type identifier.

    OTCEV_Envelope*	envelope_;
				// The envelope containing the received
				// event.

    int			error_;
    				// The error code.

    OTC_String		description_;
    				// The description of the error.

    OTC_String		origin_;
    				// The origin of the failure.

    OTC_String		details_;
    				// More details relating to the error.
};

/* ------------------------------------------------------------------------- */

#endif /* OTC_DISPATCH_FAILURE_HH */
