#ifndef OTC_INTERNET_XMLRPCSL_HH
#define OTC_INTERNET_XMLRPCSL_HH
/*
// ============================================================================
//
// = LIBRARY
//     OTC
// 
// = FILENAME
//     internet/xmlrpcsl.hh
//
// = AUTHOR(S)
//     Graham Dumpleton
// 
// = COPYRIGHT
//     Copyright 2000-2001 DUMPLETON SOFTWARE CONSULTING PTY LIMITED
//
// ============================================================================
*/

#include <OTC/internet/httpsrvl.hh>
#include <OTC/dispatch/svbroker.hh>

#ifdef __GNUG__
#pragma interface "OTC/internet/xmlrpcsl.hh"
#endif

/* ------------------------------------------------------------------------- */

class OSE_EXPORT OTC_XmlRpcServlet : public OTC_HttpServlet, public OTC_EVAgent
    // = TITLE
    //      Servlet translating RPC requests into service requests.
    //
    // = CLASS TYPE
    //      Concrete
    //
    // = DESCRIPTION
    //     The <OTC_XmlRpcServlet> class is a HTTP servlet which acts as a
    //     gateway, translating RPC requests into service requests which
    //     are then delivered to a service. Any response form the service is
    //     subsequently translated back into the appropriate response for
    //     the RPC and returned to the HTTP client. At this time only an
    //     XML-RPC style request is implemented.
{
  public:

    			~OTC_XmlRpcServlet();

    // = INITIALISATION

    			OTC_XmlRpcServlet(
			 OTC_HttpSession* theSession,
			 OTC_ServiceBinding* theBinding=0
			);
				// Creates a HTTP servlet to handle an
				// RPC request. <theSession> is the HTTP
				// session object which has parsed the HTTP
				// request. This must be supplied by the
				// factory method contained in the HTTP
				// server object. <theBinding> is the service
				// binding of the service to which the
				// RPC request should be passed. If
				// <theBinding> is <0>, a HTTP response will
				// instead be sent back indicating the
				// target URL was invalid.

  protected:

    // = REQUEST PROCESSING

    void		processRequest();
    				// Decodes the request and either flags an
				// error.

    void		processContent(OTC_String const& theContent);
    				// Accumulates the content of the request
				// and when complete forwards request
				// onto the appropriate service.

    // = TARGET SERVICE

    virtual OTC_ServiceBinding*	service(char const* theName);
				// This function is be called by the method
				// <processRequest()> with <theName> equal to
				// the servlet path supplied to the
				// constructor. This method should analyse
				// <theName> to determine what service the
				// request is being targeted at. By default
				// the method simply looks up the service
				// registry and uses the first service having
				// <theName> that it finds. You can overload
				// this method in a derived class to use
				// some other means to identify which service
				// to use. This would be appropriate given
				// that not all services may be able to
				// respond to requests conforming to what
				// can be represented in an RPC. The instance
				// of <OTC_ServiceBinding> returned by a
				// derived class must have no owner, or if it
				// has, the derived method must call <clone()>
				// on it. This is because <processRequest()>
				// will call <destroy()> on the returned
				// object when it has finished with it.

    // = EVENT HANDLING

    void		process(OTC_Event* theEvent);
    				// Deals with the response received from a
				// service, translating it back into the
				// appropriate RPC format and returning it.

    void		handle(OTC_Event* theEvent);
    				// Calls <process()> then destroys <theEvent>.

  private:

    static OTC_Boolean	decodeRequest_(
			 istream& theStream,
			 OTC_String& theMethod,
			 OTC_RWPayload& theObject
			);
    				// Decodes the actual request converting
				// int into a service request.

    static void		encodeFailure_(
			 ostream& theStream,
			 int theError,
			 char const* theDescription="",
			 char const* theOrigin="",
			 char const* theDetails=""
			);
    				// Encodes an XML-RPC failure response.

    void		sendFailure_(
			 int theError,
			 char const* theDescription="",
			 char const* theOrigin="",
			 char const* theDetails=""
			);
    				// Sends back an XML-RPC failure response.

    static OTC_Boolean	encodeRequest_(
			 ostream& theStream,
			 char const* theMethod,
			 OTC_ROPayload const& theObject
			);
    				// Encodes an XML-RPC request. Not used.

    static OTC_Boolean	encodeResponse_(
			 ostream& theStream,
			 OTC_ROPayload const& theObject
			);
    				// Encodes an XML-RPC response.

    static OTC_Boolean	encodeObject_(
			 ostream& theStream,
			 OTC_ROPayload const& theObject
			);
    				// Converts a service response into an
				// XML-RPC value.

  private:

    static OTC_Boolean	skipProlog_(istream& theStream);
    				// Skips the XML document prolog.

    static OTC_Boolean	skipProperty_(istream& theStream);
    				// Skips an XML element attribute.

    static OTC_Boolean	decodeValue_(
			 istream& theStream,
			 OTC_RWPayload theObject
			);
    				// Decodes an XML-RPC value.

    // All functions read from <theStream> with data returned via
    // <theResult>. The function returns <OTCLIB_TRUE> if successful and
    // <OTCLIB_FALSE> otherwise. A value of <OTCLIB_FALSE> will only be
    // returned when an eof condition is encountered if the function was
    // expecting to read up until some delimiter. The stream is not
    // nessarily checked for failure conditions in all circumstances.
    // If expecting to read up to a specified delimiter, a failure of
    // the stream will be noticed and <OTCLIB_FALSE> returned. In the
    // case where a delimiter is not specified, a failure of the stream
    // will simply stop the reading but not result in <OTCLIB_FALSE>
    // being returned.

    static OTC_Boolean	readName_(
			 istream& theStream,
			 OTC_String& theResult
			);
    				// Reads in a name.

    static OTC_Boolean	readDigits_(
			 istream& theStream,
			 OTC_String& theResult
			);
    				// Reads in string of decimal digits.

    static OTC_Boolean	readHexDigits_(
			 istream& theStream,
			 OTC_String& theResult
			);
    				// Reads in string of hexadecimal digits.

    // When decoding input up until some delimiter, the delimiter is always
    // left in the stream input.

    static OTC_Boolean	readToChar_(
			 istream& theStream,
			 char theChar,
			 OTC_String& theResult
			);
				// Reads in characters without any decoding
				// until <theChar> is encountered in the
				// input stream.

    static OTC_Boolean	readToString_(
			 istream& theStream,
			 char const* theString,
			 OTC_String& theResult
			);
				// Reads in characters without any decoding
				// until <theString> is encountered in the
				// input stream. Will immediately return
				// <OTCLIB_FALSE> if <theString> is <0> or
				// an empty string. It isn't possible for
				// the delimiter string to contain a null
				// character.

    static OTC_Boolean	decodeToChar_(
			 istream& theStream,
			 char theChar,
			 OTC_String& theResult
			);
				// Reads in characters until <theChar> is
				// encountered, decoding character and
				// entity references in the process. Note
				// that only inbuilt entity references are
				// supported.

    // Following wrap up standard XML and b64 encoding/decoding functions.
    // The functions for encoding strings will also ensure that any not
    // ascii text is encoded as well.

    static OTC_Boolean	decodeString_(
			 istream& theStream,
			 int theDelim,
			 OTC_String& theResult
			)
      				{
				  return xmlDecode(
				   theStream,theDelim,theResult);
				}
				// Reads characters from <theStream> undoing
				// any XML escapes using "&" escape sequence
				// until <theDelim> is encountered.
				// <theDelim> will be left in the stream. The
				// result is appended to <theResult> with the
				// function returning <OTCLIB_TRUE> if the
				// conversion was successful.

    static void		encodeString_(
			 ostream& theStream,
			 char const* theString,
			 u_int theLength
			);
				// Outputs the first <theLength> characters
				// of <theString> onto <theStream>. The
				// characters are escaped as per the XML
				// requirements for character data using
				// "&" escape sequences. Also, any non
				// ascii characters are also escaped.

    static void		encodeString_(
			 ostream& theStream,
			 char const* theString
			)
      				{
				  encodeString_(theStream,theString,
				   theString?strlen(theString):0);
				}
				// Outputs <theString> onto <theStream>. The
				// characters are escaped as per the XML
				// requirements for character data using
				// "&" escape sequences. Also, any non
				// ascii characters are also escaped.

    static OTC_Boolean	decodeBinary_(
			 istream& theStream,
			 int theDelim,
			 OTC_String& theResult
			)
      				{
				  return b64Decode(
				   theStream,theDelim,theResult);
				}
				// Reads characters from <theStream> undoing
				// any base64 encoding until <theDelim> is
				// encountered. <theDelim> will be left in
				// the stream. The result is appended to
				// <theResult> with the function returning
				// <OTCLIB_TRUE> if the conversion was
				// successful.

    static void		encodeBinary_(
			 ostream& theStream,
			 char const* theString,
			 u_int theLength
			)
      				{ b64Encode(theStream,theString,theLength); }
				// Outputs the first <theLength> characters
				// of <theString> onto <theStream>. The
				// characters are encoded in base64 format.

  private:

    			OTC_XmlRpcServlet(OTC_XmlRpcServlet const&);
				// Do not define an implementation for this.

    OTC_XmlRpcServlet&	operator=(OTC_XmlRpcServlet const&);
				// Do not define an implementation for this.

    OTC_ServiceBinding*	binding_;
    				// Service binding of target service.

    OTC_SVBroker*	broker_;
    				// Broker through which requests are sent
				// and subsequently received.

    OTC_String		content_;
    				// Accumulated content of the request.

    OTC_String		conversationId_;
    				// Conversation ID of request.
};

/* ------------------------------------------------------------------------- */

#endif /* OTC_INTERNET_XMLRPCSL_HH */
