/*
// ============================================================================
//
// = LIBRARY
//     OTC
// 
// = FILENAME
//     internet/otcxmlrpcsl.cc
//
// = AUTHOR(S)
//     Graham Dumpleton
// 
// = COPYRIGHT
//     Copyright 2000-2001 DUMPLETON SOFTWARE CONSULTING PTY LIMITED
//
// ============================================================================
*/

#ifdef __GNUG__
#pragma implementation "OTC/internet/xmlrpcsl.hh"
#endif

#include <OTC/internet/xmlrpcsl.hh>

#include <OTC/dispatch/response.hh>
#include <OTC/dispatch/failure.hh>
#include <OTC/text/osstream.hh>
#include <OTC/collctn/cbitset.hh>
#include <OTC/debug/tracer.hh>
#include <OTC/debug/tracetag.hh>

#include <OTC/stdlib/strstream.h>

#include <string.h>
#include <ctype.h>

/* ------------------------------------------------------------------------- */
static OTC_NRMutex MODULE_TRACETAG_MUTEX;
static OTC_TraceTag& MODULE_TRACETAG()
{
  static OTC_TraceTag* TAG = 0;
  MODULE_TRACETAG_MUTEX.lock();
  if (TAG == 0)
    TAG = new OTC_TraceTag("OTC_XmlRpcServlet");
  MODULE_TRACETAG_MUTEX.unlock();
  return *TAG;
}

#define MODULE MODULE_TRACETAG()

/* ------------------------------------------------------------------------- */
OTC_XmlRpcServlet::~OTC_XmlRpcServlet()
{
  if (broker_ != 0)
    delete broker_;

  if (binding_ != 0)
    binding_->destroy();
}

/* ------------------------------------------------------------------------- */
OTC_XmlRpcServlet::OTC_XmlRpcServlet(
 OTC_HttpSession* theSession,
 OTC_ServiceBinding* theBinding
)
  : OTC_HttpServlet(theSession), binding_(0), broker_(0)
{
  OTCLIB_MARKBLOCK(MODULE,"OTC_XmlRpcServlet::OTC_XmlRpcServlet(...)");

  if (theBinding != 0)
    binding_ = theBinding->clone();
}

/* ------------------------------------------------------------------------- */
void OTC_XmlRpcServlet::processRequest()
{
  OTCLIB_MARKBLOCK(MODULE,"OTC_XmlRpcServlet::processRequest()");

  // Determine if we have a service to send
  // the request to.

  if (binding_ == 0)
  {
    sendError(404,"No such service available.");
 
    return;
  }

  // Eliminate requests we can't handle.

  if (contentType() != "text/xml")
  {
    sendError(400,"Servlet can only handle text/xml.");

    return;
  }

  if (requestMethod() != "POST")
  {
    sendError(405,"Only POST method supported by servlet.");

    return;
  }

  if (contentLength() == -1)
  {
    sendError(400,"Content length must be specified.");

    return;
  }

  // The remainder of the processing is now
  // dependent on receiving actual content.
}

/* ------------------------------------------------------------------------- */
void OTC_XmlRpcServlet::processContent(OTC_String const& theContent)
{
  // Ignore any extra content beyond that
  // specified by content length.

  if (int(content_.length()) >= contentLength())
    return;

  content_ += theContent;

  // Wait for more if don't yet have what
  // is specified by content length.

  if (int(content_.length()) < contentLength())
    return;
 
  // Attempt to decode request and send it
  // on its way if everything is okay.

  OTCLIB_TRACER(MODULE) << content_ << endl << endl;

  istrstream theStream((char*)content_.string(),content_.length());

  OTC_String theMethod;
  OTC_SVPayload theObject;

  if (decodeRequest_(theStream,theMethod,theObject) == OTCLIB_FALSE)
  {
    OTCLIB_TRACER(MODULE) << "method=" << theMethod << endl << endl;
    OTCLIB_TRACER(MODULE) << theObject << endl;
    OTCLIB_TRACER(MODULE) << OTC_String::get(theStream,EOF) << endl;

    sendFailure_(54,"Request Decoding Error","netsvc.xmlrpc");

    return;
  }

  OTCLIB_TRACER(MODULE) << "method=" << theMethod << endl;
  OTCLIB_TRACER(MODULE) << theObject << endl;

  // Bind ourselves as an anonymous service
  // and send the request.

  broker_ = new OTC_SVBroker(this,"","");
  OTCLIB_ASSERT_M(broker_ != 0);

  conversationId_ = broker_->sendRequest(binding_,
   theMethod,theObject);
}

/* ------------------------------------------------------------------------- */
OTC_ServiceBinding* OTC_XmlRpcServlet::service(char const* theName)
{
  OTC_Iterator<OTC_ServiceBinding*> theBindings;
  theBindings = OTC_SVBroker::registry()->lookupServiceName(theName);

  if (theBindings.isValid())
    return theBindings.item()->clone();

  return 0;
}

/* ------------------------------------------------------------------------- */
void OTC_XmlRpcServlet::process(OTC_Event* theEvent)
{
  OTCLIB_MARKBLOCK(MODULE,"OTC_XmlRpcServlet::process(OTC_Event*)");

  if (theEvent == 0)
    return;

  OTCLIB_TRACER(MODULE) << *theEvent << endl;

  if (theEvent->type() == OTCEV_Response::typeId())
  {
    OTCEV_Response* theResponse;
    theResponse = (OTCEV_Response*)theEvent;

    if (theResponse->conversationId() == conversationId_)
    {
      OTC_ROPayload theObject;

      theObject = theResponse->content();

      if (theObject.isValid())
      {
	OTC_String theString;
	OTC_OSStream theStream(theString);

	if (encodeResponse_(theStream,theObject) != OTCLIB_FALSE)
	{
	  sendResponse(200);

	  sendHeader("Content-Type","text/xml");
	  sendHeader("Content-Length",theString.length());

	  endHeaders();

	  sendContent(theString);

	  endContent();
	}
	else
	  sendFailure_(52,"Response Encoding Error","netsvc.xmlrpc");
      }
      else
	sendFailure_(52,"Response Encoding Error","netsvc.xmlrpc");
    }
  }
  else if (theEvent->type() == OTCEV_Failure::typeId())
  {
    OTCEV_Failure* theFailure;
    theFailure = (OTCEV_Failure*)theEvent;

    if (theFailure->conversationId() == conversationId_)
    {
      OTC_String theString;

      OTC_OSStream theStream(theString);

      encodeFailure_(theStream,theFailure->error(),
       theFailure->description(),theFailure->origin(),
       theFailure->details());

      sendResponse(200);

      sendHeader("Content-Type","text/xml");
      sendHeader("Content-Length",theString.length());

      endHeaders();

      sendContent(theString);

      endContent();
    }
  }
}

/* ------------------------------------------------------------------------- */
void OTC_XmlRpcServlet::handle(OTC_Event* theEvent)
{
  if (theEvent == 0)
    return;

  process(theEvent);

  theEvent->destroy();
}

/* ------------------------------------------------------------------------- */
OTC_Boolean OTC_XmlRpcServlet::decodeRequest_(
 istream& theStream,
 OTC_String& theMethod,
 OTC_RWPayload& theObject
)
{
  theMethod.truncate();

  // Ensure that payload starts out as
  // an empty array so that empty array
  // gets passed back if no params.

  theObject <<= OTC_ROPayload::nullArray();

  // First find and skip xml document
  // header.

  theStream >> ws;

  if (!skipProlog_(theStream))
    return OTCLIB_FALSE;

  theStream >> ws;

  // Verify we have a request and work
  // out what the method name is.

  OTC_String theString;

  if (theStream.get() != '<')
    return OTCLIB_FALSE;

  if (!readName_(theStream,theString))
    return OTCLIB_FALSE;

  if (theString != "methodCall")
    return OTCLIB_FALSE;

  theStream >> ws;

  if (theStream.get() != '>')
    return OTCLIB_FALSE;

  theStream >> ws;

  if (theStream.get() != '<')
    return OTCLIB_FALSE;

  if (!readName_(theStream,theString))
    return OTCLIB_FALSE;

  if (theString != "methodName")
    return OTCLIB_FALSE;

  theStream >> ws;

  // Check for possibility of <methodName/>.

  if (theStream.peek() == '/')
  {
    // Can only be <methodName/>.

    theStream.ignore(1);

    if (theStream.get() != '>')
      return OTCLIB_FALSE;
  }
  else
  {
    // Can only be <methodName>.

    if (theStream.get() != '>')
      return OTCLIB_FALSE;

    if (!decodeString_(theStream,'<',theMethod))
      return OTCLIB_FALSE;

    if (theStream.get() != '<')
      return OTCLIB_FALSE;

    if (theStream.get() != '/')
      return OTCLIB_FALSE;

    if (!readName_(theStream,theString))
      return OTCLIB_FALSE;

    if (theString != "methodName")
      return OTCLIB_FALSE;

    theStream >> ws;

    if (theStream.get() != '>')
      return OTCLIB_FALSE;
  }

  // We now have to decode the params.
  // We have to cope with the fact that
  // there may be no <params>, that it
  // is empty and where it is empty that
  // it may use either <params></params>
  // or <params/> forms.

  theStream >> ws;

  if (theStream.get() != '<')
    return OTCLIB_FALSE;

  // Check here for possibility of having
  // </methodCall>.

  if (theStream.peek() == '/')
  {
    // No <params> possible, must be the
    // </methodCall>.

    theStream.ignore(1);

    if (!readName_(theStream,theString))
      return OTCLIB_FALSE;

    if (theString != "methodCall")
      return OTCLIB_FALSE;

    theStream >> ws;

    if (theStream.get() != '>')
      return OTCLIB_FALSE;

    theStream >> ws;

    if (!theStream.eof())
      return OTCLIB_FALSE;

    return OTCLIB_TRUE;
  }

  // Can only be <params> or <params/>.

  if (!readName_(theStream,theString))
    return OTCLIB_FALSE;

  if (theString != "params")
    return OTCLIB_FALSE;

  theStream >> ws;

  // Now have to check for whether
  // this might be <params/>.

  if (theStream.peek() == '/')
  {
    // Only possibility is <params/>.

    theStream.ignore(1);

    if (theStream.get() != '>')
      return OTCLIB_FALSE;
  }
  else
  {
    // Only possibility is <params>.

    if (theStream.get() != '>')
      return OTCLIB_FALSE;

    theStream >> ws;

    if (theStream.get() != '<')
      return OTCLIB_FALSE;

    // Now have to check whether this
    // might be </params>.

    if (theStream.peek() == '/')
    {
      // Only possibility is </params>.

      theStream.ignore(1);

      if (!readName_(theStream,theString))
	return OTCLIB_FALSE;

      if (theString != "params")
	return OTCLIB_FALSE;

      theStream >> ws;

      if (theStream.get() != '>')
	return OTCLIB_FALSE;
    }
    else
    {
      // Only possibility is <param>. Note that
      // not accepting <param/> as being valid as
      // should contain at least <value/>.

      if (!readName_(theStream,theString))
	return OTCLIB_FALSE;

      if (theString != "param")
	return OTCLIB_FALSE;

      theStream >> ws;

      if (theStream.get() != '>')
	return OTCLIB_FALSE;

      u_int theIndex = 0;

      while (1)
      {
	theStream >> ws;

	if (theStream.get() != '<')
	  return OTCLIB_FALSE;

	if (!readName_(theStream,theString))
	  return OTCLIB_FALSE;

	if (theString != "value")
	  return OTCLIB_FALSE;

	theStream >> ws;

	// Have to check whether this might
	// might be <value/>.

	if (theStream.peek() == '/')
	{
	  // Can only be <value/>.

	  theStream.ignore(1);

	  if (theStream.get() != '>')
	    return OTCLIB_FALSE;

	  theObject[theIndex++] <<= "";
	}
	else
	{
	  // Can only be <value>.

	  if (theStream.get() != '>')
	    return OTCLIB_FALSE;

	  if (!decodeValue_(theStream,theObject[theIndex++]))
	    return OTCLIB_FALSE;
	}

	theStream >> ws;

	if (theStream.get() != '<')
	  return OTCLIB_FALSE;

	if (theStream.get() != '/')
	  return OTCLIB_FALSE;

	if (!readName_(theStream,theString))
	  return OTCLIB_FALSE;

	if (theString != "param")
	  return OTCLIB_FALSE;

	theStream >> ws;

	if (theStream.get() != '>')
	  return OTCLIB_FALSE;

	theStream >> ws;

	if (theStream.get() != '<')
	  return OTCLIB_FALSE;

	if (theStream.peek() == '/')
	  break;

	if (!readName_(theStream,theString))
	  return OTCLIB_FALSE;

	if (theString != "param")
	  return OTCLIB_FALSE;

	theStream >> ws;

	if (theStream.get() != '>')
	  return OTCLIB_FALSE;
      }

      theStream.ignore(1);

      if (!readName_(theStream,theString))
	return OTCLIB_FALSE;

      if (theString != "params")
	return OTCLIB_FALSE;

      theStream >> ws;

      if (theStream.get() != '>')
	return OTCLIB_FALSE;
    }
  }

  // Make sure method call is closed
  // off correctly.

  theStream >> ws;

  if (theStream.get() != '<')
    return OTCLIB_FALSE;

  if (theStream.peek() != '/')
    return OTCLIB_FALSE;

  theStream.ignore(1);

  if (!readName_(theStream,theString))
    return OTCLIB_FALSE;

  if (theString != "methodCall")
    return OTCLIB_FALSE;

  theStream >> ws;

  if (theStream.get() != '>')
    return OTCLIB_FALSE;

  theStream >> ws;

  if (!theStream.eof())
    return OTCLIB_FALSE;

  return OTCLIB_TRUE;
}

/* ------------------------------------------------------------------------- */
void OTC_XmlRpcServlet::encodeFailure_(
 ostream& theStream,
 int theError,
 char const* theDescription,
 char const* theOrigin,
 char const* theDetails
)
{
  theStream << "<?xml version=\"1.0\"?>\n";
  theStream << "<methodResponse>\n";
  theStream << "<fault>\n";
  theStream << "<value>\n";
  theStream << "<struct>\n";
  theStream << "<member>\n";
  theStream << "<name>faultCode</name>\n";
  theStream << "<value><int>";
  theStream << theError;
  theStream << "</int></value>\n";
  theStream << "</member>\n";
  theStream << "<member>\n";
  theStream << "<name>faultString</name>\n";
  theStream << "<value>";

  if (theOrigin != 0 && *theOrigin != EOS)
  {
    encodeString_(theStream,theOrigin);
    theStream << " -- ";
  }

  encodeString_(theStream,theDescription);

  if (theDetails != 0 && *theDetails != EOS)
  {
    theStream << "\n\n";
    encodeString_(theStream,theDetails);
  }

  theStream << "</value>\n";
  theStream << "</member>\n";
  theStream << "</struct>\n";
  theStream << "</value>\n";
  theStream << "</fault>\n";
  theStream << "</methodResponse>\n";
  theStream << flush;
}

/* ------------------------------------------------------------------------- */
void OTC_XmlRpcServlet::sendFailure_(
 int theError,
 char const* theDescription,
 char const* theOrigin,
 char const* theDetails
)
{
  OTC_String theString;

  OTC_OSStream theStream(theString);

  encodeFailure_(theStream,theError,
   theDescription,theOrigin,theDetails);

  sendResponse(200);

  sendHeader("Content-Type","text/xml");
  sendHeader("Content-Length",theString.length());

  endHeaders();

  sendContent(theString);

  endContent();
}

/* ------------------------------------------------------------------------- */
OTC_Boolean OTC_XmlRpcServlet::encodeRequest_(
 ostream& theStream,
 char const* theMethod,
 OTC_ROPayload const& theObject
)
{
  theStream << "<?xml version=\"1.0\"?>\n";
  theStream << "<methodCall>\n";
  theStream << "<params>\n";

  if (theObject.nodeType() != "array")
    return OTCLIB_FALSE;

  OTC_ROPayload theReader;
  theReader = theObject.firstChild();

  while (theReader.isValid())
  {
    theStream << "<param>\n";

    if (!encodeObject_(theStream,theReader))
      return OTCLIB_FALSE;

    theStream << "</param>\n";

    theReader = theReader.nextSibling();
  }

  theStream << "</params>\n";
  theStream << "</methodCall>\n";
  theStream << flush;

  return OTCLIB_TRUE;
}

/* ------------------------------------------------------------------------- */
OTC_Boolean OTC_XmlRpcServlet::encodeResponse_(
 ostream& theStream,
 OTC_ROPayload const& theObject
)
{
  theStream << "<?xml version=\"1.0\"?>\n";
  theStream << "<methodResponse>\n";
  theStream << "<params>\n";
  theStream << "<param>\n";

  if (!encodeObject_(theStream,theObject))
    return OTCLIB_FALSE;

  theStream << "</param>\n";
  theStream << "</params>\n";
  theStream << "</methodResponse>\n";
  theStream << flush;

  return OTCLIB_TRUE;
}

/* ------------------------------------------------------------------------- */
OTC_Boolean OTC_XmlRpcServlet::encodeObject_(
 ostream& theStream,
 OTC_ROPayload const& theObject
)
{
  theStream << "<value>";

  if (theObject.nodeType() == "scalar")
  {
    if (theObject.property("type") == "xsd:integer" ||
     theObject.property("type") == "xsd:long" ||
     theObject.property("type") == "xsd:int" ||
     theObject.property("type") == "xsd:short" ||
     theObject.property("type") == "xsd:byte" ||
     theObject.property("type") == "xsd:unsignedLong" ||
     theObject.property("type") == "xsd:unsignedInt" ||
     theObject.property("type") == "xsd:unsignedShort" ||
     theObject.property("type") == "xsd:unsignedByte")
    {
      theStream << "<int>";
      encodeString_(theStream,theObject.value().string(),
       theObject.value().length());
      theStream << "</int>";
    }
    else if (theObject.property("type") == "xsd:double" ||
     theObject.property("type") == "xsd:float")
    {
      theStream << "<double>";
      encodeString_(theStream,theObject.value().string(),
       theObject.value().length());
      theStream << "</double>";
    }
    else if (theObject.property("type") == "xsd:boolean")
    {
      theStream << "<boolean>";
      if (theObject.value() == "false" || theObject.value() == "0")
	theStream << "0";
      else if (theObject.value() == "true" || theObject.value() == "1")
	theStream << "1";
      else
	return OTCLIB_FALSE;

      theStream << "</boolean>";
    }
    else if (theObject.property("type") == "xsd:base64Binary")
    {
      theStream << "<base64>";
      encodeBinary_(theStream,theObject.value().string(),
       theObject.value().length());
      theStream << "</base64>";
    }
    else if (theObject.property("type") == "xsd:dateTime")
    {
      theStream << "<dateTime.iso8601>";

      if (OTC_Time::isValidTime(theObject.value().string()))
      {
	OTC_Time theTime(theObject.value().string());
	OTC_String theString = theTime.asString("%Y%m%dT%H:%M:%S");

	encodeString_(theStream,theString.string(),theString.length());
      }
      else
	return OTCLIB_FALSE;

      theStream << "</dateTime.iso8601>";
    }
    // else if (theObject.property("type") == "")
    // {
    //   theStream << "<nil/>";
    // }
    else
    {
      theStream << "<string>";
      encodeString_(theStream,theObject.value().string(),
       theObject.value().length());
      theStream << "</string>";
    }
  }
  else if (theObject.nodeType() == "array")
  {
    theStream << "<array>\n";
    theStream << "<data>\n";

    OTC_ROPayload theReader;
    theReader = theObject.firstChild();

    while (theReader.isValid())
    {
      if (!encodeObject_(theStream,theReader))
	return OTCLIB_FALSE;

      theReader = theReader.nextSibling();
    }

    theStream << "</data>\n";
    theStream << "</array>";
  }
  else if (theObject.nodeType() == "index")
  {
    theStream << "<struct>\n";

    OTC_ROPayload theReader;
    theReader = theObject.firstChild();

    while (theReader.isValid())
    {
      theStream << "<member>\n";
      theStream << "<name>";
      encodeString_(theStream,
       theReader.nodeName().string(),
       theReader.nodeName().length());
      theStream << "</name>\n";

      if (!encodeObject_(theStream,theReader))
	return OTCLIB_FALSE;

      theStream << "</member>\n";

      theReader = theReader.nextSibling();
    }

    theStream << "</struct>";
  }
  else
    return OTCLIB_FALSE;

  theStream << "</value>\n";

  return OTCLIB_TRUE;
}

/* ------------------------------------------------------------------------- */
OTC_Boolean OTC_XmlRpcServlet::skipProlog_(istream& theStream)
{
  theStream >> ws;

  if (theStream.get() != '<')
    return OTCLIB_FALSE;

  if (theStream.get() != '?')
    return OTCLIB_FALSE;

  OTC_String tmpString;

  if (!readName_(theStream,tmpString))
    return OTCLIB_FALSE;

  if (!tmpString.compare("xml",OTCLIB_IGNORECASE))
    return OTCLIB_FALSE;

  // XXX Should read all version info etc.

#if 0
  theStream >> ws;

  while (theStream.peek() != '?')
  {
     if (skipProperty_(theStream) == OTCLIB_FALSE)
       return OTCLIB_FALSE;

     theStream >> ws;
  }

  theStream.ignore(1);

  if (theStream.get() != '>')
    return OTCLIB_FALSE;
#endif

  if (!readToString_(theStream,"?>",tmpString))
    return OTCLIB_FALSE;

  return OTCLIB_TRUE;
}

/* ------------------------------------------------------------------------- */
OTC_Boolean OTC_XmlRpcServlet::skipProperty_(istream& theStream)
{
  OTC_String theName;
  theName.readToDelim(theStream,'=');

  if (theStream.fail())
    return OTCLIB_FALSE;

  theStream >> ws;

  OTC_String theValue;

  if (theStream.peek() == '"')
  {
    theStream.ignore(1);

    if (decodeString_(theStream,'"',theValue) == OTCLIB_FALSE)
      return OTCLIB_FALSE;
  }
  else if (theStream.peek() == '\'')
  {
    theStream.ignore(1);

    if (decodeString_(theStream,'\'',theValue) == OTCLIB_FALSE)
      return OTCLIB_FALSE;
  }
  else
    return OTCLIB_FALSE;

  theStream.ignore(1);

  return OTCLIB_TRUE;
}

/* ------------------------------------------------------------------------- */
OTC_Boolean OTC_XmlRpcServlet::decodeValue_(
 istream& theStream,
 OTC_RWPayload theObject
)
{
  OTC_String theString;

  // Have to deal with case whereby
  // may be value with no type.

  OTC_String theContent;
  OTC_String theElement;

  if (!decodeString_(theStream,'<',theContent))
    return OTCLIB_FALSE;

  if (theStream.get() != '<')
    return OTCLIB_FALSE;

  // Check for possibility of </value>.

  if (theStream.peek() == '/')
  {
    // Can only be </value>.

    theStream.ignore(1);

    if (!readName_(theStream,theString))
      return OTCLIB_FALSE;

    if (theString != "value")
      return OTCLIB_FALSE;

    theStream >> ws;

    if (theStream.get() != '>')
      return OTCLIB_FALSE;

    theObject.assign(theContent,theContent.length());

    return OTCLIB_TRUE;
  }
  else
  {
    if (!readName_(theStream,theElement))
      return OTCLIB_FALSE;

    theStream >> ws;

    // Check for possibility of an empty
    // tag.

    if (theStream.peek() == '/')
    {
      // Must be an empty tag.

      theStream.ignore(1);

      if (theStream.get() != '>')
	return OTCLIB_FALSE;

      if (theElement == "struct")
	theObject <<= OTC_ROPayload::nullIndex();
      else if (theElement == "i4" || theElement == "int")
	theObject.assign("","xsd:int");
      else if (theElement == "boolean")
	theObject.assign("","xsd:boolean");
      else if (theElement == "double")
	theObject.assign("","xsd:double");
      else if (theElement == "dateTime.iso8601")
	theObject.assign("","xsd:dateTime");
      else if (theElement == "base64")
	theObject.assign("","xsd:base64Binary");
      else if (theElement == "string")
	theObject.assign("","xsd:string");
      // else if (theElement == "nil")
      //   theObject <<= OTC_ROPayload::nullValue();
      else
	return OTCLIB_FALSE;
    }
    else
    {
      if (theStream.get() != '>')
	return OTCLIB_FALSE;

      if (theElement == "array")
      {
	theObject <<= OTC_ROPayload::nullArray();

	theStream >> ws;

	if (theStream.get() != '<')
	  return OTCLIB_FALSE;

	if (!readName_(theStream,theString))
	  return OTCLIB_FALSE;

	if (theString != "data")
	  return OTCLIB_FALSE;

	theStream >> ws;

	// Check for possibility of <data/>.

	if (theStream.peek() == '/')
	{
	  // Can only be <data/>

	  theStream.ignore(1);

	  if (theStream.get() != '>')
	    return OTCLIB_FALSE;
	}
	else
	{
	  // Can only be <data>.

	  if (theStream.get() != '>')
	    return OTCLIB_FALSE;

	  theStream >> ws;

	  if (theStream.get() != '<')
	    return OTCLIB_FALSE;

	  while (1)
	  {
	    // Check for possibility of </data>.

	    if (theStream.peek() == '/')
	    {
	      // Can only be </data>.

	      theStream.ignore(1);

	      if (!readName_(theStream,theString))
		return OTCLIB_FALSE;

	      if (theString != "data")
		return OTCLIB_FALSE;

	      theStream >> ws;

	      if (theStream.get() != '>')
		return OTCLIB_FALSE;

	      break;
	    }
	    else
	    {
	      // Can only be <value> or <value/>.

	      if (!readName_(theStream,theString))
		return OTCLIB_FALSE;

	      if (theString != "value")
		return OTCLIB_FALSE;

	      theStream >> ws;

	      // Check for possibility of <value/>.

	      if (theStream.peek() == '/')
	      {
		// Can only be <value/>.

		theStream.ignore(1);

		if (theStream.get() != '>')
		  return OTCLIB_FALSE;

		theObject[-1] <<= "";
	      }
	      else
	      {
		// Can only be <value>.

		if (theStream.get() != '>')
		  return OTCLIB_FALSE;

		if (!decodeValue_(theStream,theObject[-1]))
		  return OTCLIB_FALSE;
	      }
	    }

	    theStream >> ws;

	    if (theStream.get() != '<')
	      return OTCLIB_FALSE;
	  }
	}

	theStream >> ws;

	if (theStream.get() != '<')
	  return OTCLIB_FALSE;

	theString.readToDelim(theStream,'>');
	theString.rtrim();

	if (theString != "/array")
	  return OTCLIB_FALSE;
      }
      else if (theElement == "struct")
      {
	theStream >> ws;

	if (theStream.get() != '<')
	  return OTCLIB_FALSE;

	theString.readToDelim(theStream,'>');
	theString.rtrim();

	while (theString == "member")
	{
	  theStream >> ws;

	  if (theStream.get() != '<')
	    return OTCLIB_FALSE;

	  if (!readName_(theStream,theString))
	    return OTCLIB_FALSE;

	  if (theString != "name")
	    return OTCLIB_FALSE;

	  theStream >> ws;

	  OTC_String theMember;

	  // Check for possibility of <name/>

	  if (theStream.peek() == '/')
	  {
	    // Can only be <name/>.

	    theStream.ignore(1);

	    if (theStream.get() != '>')
	      return OTCLIB_FALSE;
	  }
	  else
	  {
	    // Can only be <name>.

	    if (theStream.get() != '>')
	      return OTCLIB_FALSE;

	    if (!decodeString_(theStream,'<',theMember))
	      return OTCLIB_FALSE;

	    theStream >> ws;

	    if (theStream.get() != '<')
	      return OTCLIB_FALSE;

	    theString.readToDelim(theStream,'>');
	    theString.rtrim();

	    if (theString != "/name")
	      return OTCLIB_FALSE;
	  }

	  theStream >> ws;

	  if (theStream.get() != '<')
	    return OTCLIB_FALSE;

	  if (!readName_(theStream,theString))
	    return OTCLIB_FALSE;

	  if (theString != "value")
	    return OTCLIB_FALSE;

	  theStream >> ws;

	  // Check for possibility of <value/>.

	  if (theStream.peek() == '/')
	  {
	    // Can only be <value/>.

	    theStream.ignore(1);

	    if (theStream.get() != '>')
	      return OTCLIB_FALSE;

	    theObject[theMember] <<= "";
	  }
	  else
	  {
	    // Can only be <value>.

	    if (theStream.get() != '>')
	      return OTCLIB_FALSE;

	    if (!decodeValue_(theStream,theObject[theMember]))
	      return OTCLIB_FALSE;
	  }

	  theStream >> ws;

	  if (theStream.get() != '<')
	    return OTCLIB_FALSE;

	  theString.readToDelim(theStream,'>');
	  theString.rtrim();

	  if (theString != "/member")
	    return OTCLIB_FALSE;

	  theStream >> ws;

	  if (theStream.get() != '<')
	    return OTCLIB_FALSE;

	  theString.readToDelim(theStream,'>');
	  theString.rtrim();
	}

	if (theString != "/struct")
	  return OTCLIB_FALSE;
      }
      else
      {
	theContent.truncate();

	if (theElement == "base64")
	{
	  if (!decodeBinary_(theStream,'<',theContent))
	    return OTCLIB_FALSE;
	}
	else if (theElement == "dateTime.iso8601")
	{
	  if (!decodeString_(theStream,'<',theContent))
	    return OTCLIB_FALSE;

	  u_int theYear;
	  u_int theMonth;
	  u_int theDay;
	  u_int theHour;
	  u_int theMin;
	  u_int theSec;

	  char theNext;

	  if (theContent.index(EOS) != -1)
	    return OTCLIB_FALSE;

	  if (sscanf(theContent,"%4u%2u%2uT%2u:%2u:%2u%c",
	   &theYear,&theMonth,&theDay,&theHour,&theMin,&theSec,&theNext) == 6
	  )
	  {
	    char buf[64];

	    sprintf(buf,"%04u-%02u-%02uT%02u:%02u:%02u",
	     theYear,theMonth,theDay,theHour,theMin,theSec);

	    theContent = buf;
	  }
	  else
	    return OTCLIB_FALSE;
	}
	else
	{
	  if (!decodeString_(theStream,'<',theContent))
	    return OTCLIB_FALSE;
	}

	if (theElement == "boolean")
	{
	  if (theContent == "0")
	    theObject.assign("false");
	  else
	    theObject.assign("true");
	}
	else
	  theObject.assign(theContent,theContent.length());

	if (theElement == "i4" || theElement == "int")
	  theObject.setProperty("type","xsd:int");
	else if (theElement == "boolean")
	  theObject.setProperty("type","xsd:boolean");
	else if (theElement == "double")
	  theObject.setProperty("type","xsd:double");
	else if (theElement == "dateTime.iso8601")
	  theObject.setProperty("type","xsd:dateTime");
	else if (theElement == "base64")
	  theObject.setProperty("type","xsd:base64Binary");
	// else if (theElement == "nil")
	//   theObject.setProperty("type","");
	else if (theElement != "string")
	  return OTCLIB_FALSE;

	if (theStream.get() != '<')
	  return OTCLIB_FALSE;

	if (theStream.get() != '/')
	  return OTCLIB_FALSE;

	theString.readToDelim(theStream,'>');
	theString.rtrim();

	if (theString != theElement)
	  return OTCLIB_FALSE;
      }
    }
  }

  theStream >> ws;

  if (theStream.get() != '<')
    return OTCLIB_FALSE;

  theString.readToDelim(theStream,'>');
  theString.rtrim();

  if (theString != "/value")
    return OTCLIB_FALSE;

  return OTCLIB_TRUE;
}

/* ------------------------------------------------------------------------- */
OTC_Boolean OTC_XmlRpcServlet::readName_(
 istream& theStream,
 OTC_String& theResult
)
{
  theResult.truncate();

  int theInput;

  // NameChar ::= Letter | Digit | '.' | '-' | '_' | ':'
  // Name ::= (Letter | '_' | ':') (NameChar)*

  theInput = theStream.peek();

  if (theInput == EOF || (!isalpha(theInput) && !strchr("_:",theInput)))
    return OTCLIB_FALSE;

  theResult += theStream.get();

  theInput = theStream.peek();

  while (theInput != EOF && (isalnum(theInput) || strchr(".-_:",theInput)))
  {
    theResult += theStream.get();

    theInput = theStream.peek();
  }

  return OTCLIB_TRUE;
}

/* ------------------------------------------------------------------------- */
OTC_Boolean OTC_XmlRpcServlet::readDigits_(
 istream& theStream,
 OTC_String& theResult
)
{
  theResult.truncate();

  int theInput;

  // Digits ::= [0-9]+

  theInput = theStream.peek();

  if (theInput == EOF || !isdigit(theInput))
    return OTCLIB_FALSE;

  theResult += theStream.get();

  theInput = theStream.peek();

  while (theInput != EOF && isdigit(theInput))
  {
    theResult += theStream.get();

    theInput = theStream.peek();
  }

  return OTCLIB_TRUE;
}

/* ------------------------------------------------------------------------- */
OTC_Boolean OTC_XmlRpcServlet::readHexDigits_(
 istream& theStream,
 OTC_String& theResult
)
{
  theResult.truncate();

  int theInput;

  // HexDigits ::= [0-9][a-f][A-F]+

  theInput = theStream.peek();

  if (theInput == EOF || !isxdigit(theInput))
    return OTCLIB_FALSE;

  theResult += theStream.get();

  theInput = theStream.peek();

  while (theInput != EOF && isxdigit(theInput))
  {
    theResult += theStream.get();

    theInput = theStream.peek();
  }

  return OTCLIB_TRUE;
}

/* ------------------------------------------------------------------------- */
OTC_Boolean OTC_XmlRpcServlet::readToChar_(
 istream& theStream,
 char theChar,
 OTC_String& theResult
)
{
  theResult.truncate();

  int theInput;

  theInput = theStream.peek();

  while (theInput != EOF && theInput != theChar)
  {
    theResult += theStream.get();

    theInput = theStream.peek();
  }

  if (theInput == EOF)
    return OTCLIB_FALSE;

  return OTCLIB_TRUE;
}

/* ------------------------------------------------------------------------- */
OTC_Boolean OTC_XmlRpcServlet::readToString_(
 istream& theStream,
 char const* theString,
 OTC_String& theResult
)
{
  if (theString == 0 || *theString == EOS)
    return OTCLIB_FALSE;

  int theInput;
  u_int theLength;
  OTC_String tmpString;

  theLength = strlen(theString);

  if (theLength == 1)
    return readToChar_(theStream,*theString,theResult);

  theResult.truncate();

  theInput = theStream.peek();

  while (theInput != EOF)
  {
    if (theInput == *theString)
    {
      // We might have a match. If we
      // don't match we have to cycle
      // data through the temporary
      // string while we still might
      // have part of the data which
      // constitutes the match.

      tmpString.truncate();

      OTC_String::read(tmpString,theStream,theLength);

      if (tmpString.length() != theLength)
	return OTCLIB_FALSE;

      while (1)
      {
	if (tmpString == theString)
	  return OTCLIB_TRUE;

	theResult += tmpString[u_int(0)];
	tmpString.remove(0,1);

	while (!tmpString.isEmpty())
	{
	  theInput = tmpString[u_int(0)];

	  if (theInput == *theString)
	    break;

	  theResult += theInput;
	  tmpString.remove(0,1);
	}

	if (tmpString.isEmpty())
	  break;

	OTC_String::read(tmpString,theStream,theLength-tmpString.length());

	if (tmpString.length() != theLength)
	  return OTCLIB_FALSE;
      }
    }
    else
      theResult += theStream.get();

    theInput = theStream.peek();
  }

  if (theInput == EOF)
    return OTCLIB_FALSE;

  return OTCLIB_TRUE;
}

/* ------------------------------------------------------------------------- */
OTC_Boolean OTC_XmlRpcServlet::decodeToChar_(
 istream& theStream,
 char theChar,
 OTC_String& theResult
)
{
  theResult.truncate();

  int theInput;
  OTC_String tmpString;

  theInput = theStream.peek();

  while (theInput != EOF && theInput != theChar)
  {
    theStream.ignore(1);

    if (theInput == '&')
    {
      // This should be a character or
      // entity reference.

      // Reference ::= EntityRef | CharRef
      // CharRef ::= '&#' [0-9]+ ';' | '&#x' [0-9a-fA-F]+ ';'
      // EntityRef ::= '&' Name ';'

      theInput = theStream.peek();

      if (theInput == '#')
      {
	// Character reference.

	theStream.ignore(1);

	theInput = theStream.peek();

	if (theInput == 'x')
	{
	  // Hexadecimal.

	  if (!readHexDigits_(theStream,tmpString))
	    return OTCLIB_FALSE;

	  if (theStream.peek() != ';')
	    return OTCLIB_FALSE;

	  int v1 = 0;
	  char v2 = 0;

	  if (sscanf(tmpString,"%x%c",&v1,&v2) != 1)
	    return OTCLIB_FALSE;

	  if (v1 > 0xFF)
	    return OTCLIB_FALSE;

	  theResult += char(v1);
	}
	else if (isdigit(theInput))
	{
	  // Numeric.

	  if (!readDigits_(theStream,tmpString))
	    return OTCLIB_FALSE;

	  if (theStream.peek() != ';')
	    return OTCLIB_FALSE;

	  int v1 = 0;
	  char v2 = 0;

	  if (sscanf(tmpString,"%u%c",&v1,&v2) != 1)
	    return OTCLIB_FALSE;

	  if (v1 > 0xFF)
	    return OTCLIB_FALSE;

	  theResult += char(v1);
	}
	else
	  return OTCLIB_FALSE;
      }
      else
      {
	// Entity reference.

	if (!readName_(theStream,tmpString))
	  return OTCLIB_FALSE;

	if (theStream.peek() != ';')
	  return OTCLIB_FALSE;

	if (tmpString == "amp")
	  theResult += '&';
	else if (tmpString == "lt")
	  theResult += '<';
	else if (tmpString == "gt")
	  theResult += '>';
	else if (tmpString == "apos")
	  theResult += '\'';
	else if (tmpString == "quot")
	  theResult += '"';
	else
	  return OTCLIB_FALSE;

	theStream.ignore(1);
      }
    }
    else
      theResult += theInput;

    theInput = theStream.get();
  }

  if (theInput == EOF)
    return OTCLIB_FALSE;

  return OTCLIB_TRUE;
}

/* ------------------------------------------------------------------------- */
void OTC_XmlRpcServlet::encodeString_(
 ostream& theStream,
 char const* theString,
 u_int theLength
)
{
  static int init = 0;
  static OTC_CBitSet theChars;

  if (init == 0)
  {
    init = 1;
    theChars.set(128,128);
  }

  xmlEncode(theStream,theString,theLength,theChars);
}

/* ------------------------------------------------------------------------- */
